# Copyright 2014-2022 The PySCF Developers, 2025-2026 Sidereus-AI ltd. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");


from pyscf import scf
import numpy
from pyscf import lib
from pyscf.lib import logger
from pyscf.scf import ucphf
from pyscf.scf import _response_functions
from . import rhf as rhf_nmr


def dia(nmrobj, shielding_nuc=None, dm0=None):
    if dm0 is None:
        dm0 = nmrobj._scf.make_rdm1()
    if not (isinstance(dm0, numpy.ndarray) and dm0.ndim == 2):
        dm0 = dm0[0] + dm0[1]
    return rhf_nmr.dia(nmrobj, shielding_nuc, dm0)


def para(nmrobj, mo10=None, mo_coeff=None, mo_occ=None,
         shielding_nuc=None):
    if mo_coeff is None:
        mo_coeff = nmrobj._scf.mo_coeff
    if mo_occ is None:
        mo_occ = nmrobj._scf.mo_occ
    if shielding_nuc is None:
        shielding_nuc = nmrobj.shielding_nuc

    mol = nmrobj.mol
    para_vir = numpy.empty((len(shielding_nuc), 3, 3))
    para_occ = numpy.empty((len(shielding_nuc), 3, 3))
    occidxa = mo_occ[0] > 0
    occidxb = mo_occ[1] > 0
    viridxa = mo_occ[0] == 0
    viridxb = mo_occ[1] == 0
    orboa = mo_coeff[0][:, occidxa]
    orbob = mo_coeff[1][:, occidxb]
    orbva = mo_coeff[0][:, viridxa]
    orbvb = mo_coeff[1][:, viridxb]
    nao = mo_coeff[0].shape[0]
    # Vectorized version: extract occupied and virtual parts and perform matrix multiplication in batch
    mo10a_occ = mo10[0][:, occidxa, :]  # shape: (3, nocc_alpha, nocc_alpha)
    mo10b_occ = mo10[1][:, occidxb, :]  # shape: (3, nocc_beta, nocc_beta)
    mo10a_vir = mo10[0][:, viridxa, :]  # shape: (3, nvir_alpha, nocc_alpha)
    mo10b_vir = mo10[1][:, viridxb, :]  # shape: (3, nvir_beta, nocc_beta)

    # Compute contributions from alpha and beta electrons
    dm10_oo_alpha = orboa @ mo10a_occ @ orboa.conj().T  # shape: (3, nao, nao)
    dm10_oo_beta = orbob @ mo10b_occ @ orbob.conj().T   # shape: (3, nao, nao)
    dm10_vo_alpha = orbva @ mo10a_vir @ orboa.conj().T  # shape: (3, nao, nao)
    dm10_vo_beta = orbvb @ mo10b_vir @ orbob.conj().T   # shape: (3, nao, nao)

    # Combine contributions from both spin channels
    dm10_oo = dm10_oo_alpha + dm10_oo_beta
    dm10_vo = dm10_vo_alpha + dm10_vo_beta
    for n, atm_id in enumerate(shielding_nuc):
        mol.set_rinv_origin(mol.atom_coord(atm_id))
        h01 = mol.intor_asymmetric('int1e_prinvxp', 3)
        para_occ[n] = numpy.einsum('xji,yij->xy', dm10_oo, h01) * 2
        para_vir[n] = numpy.einsum('xji,yij->xy', dm10_vo, h01) * 2
    msc_para = para_occ + para_vir
    return msc_para, para_vir, para_occ


def make_h10(mol, dm0, verbose=logger.WARN):
    log = logger.new_logger(mol, verbose=verbose)
    # A10_i dot p + p dot A10_i consistents with <p^2 g>
    # A10_j dot p + p dot A10_j consistents with <g p^2>
    # A10_j dot p + p dot A10_j => i/2 (rjxp - pxrj) = irjxp
    log.debug('First-order GIAO Fock matrix')
    h1 = -.5 * mol.intor('int1e_giao_irjxp', 3) + make_h10giao(mol, dm0)
    return h1


def make_h10giao(mol, dm0):
    vj, vk = rhf_nmr.get_jk(mol, dm0)
    h1 = vj[0] + vj[1] - vk
    h1 -= mol.intor_asymmetric('int1e_ignuc', 3)
    if mol.has_ecp():
        h1 -= mol.intor_asymmetric('ECPscalar_ignuc', 3)
    h1 -= mol.intor('int1e_igkin', 3)
    return h1


def get_fock(nmrobj, dm0=None):
    r'''First order partial derivatives of Fock matrix wrt external magnetic
    field.  \frac{\partial F}{\partial B}
    '''
    if dm0 is None:
        dm0 = nmrobj._scf.make_rdm1()

    log = logger.Logger(nmrobj.stdout, nmrobj.verbose)
    h1 = make_h10(nmrobj.mol, dm0, log)
    if nmrobj.chkfile:
        lib.chkfile.dump(nmrobj.chkfile, 'nmr/h1', h1)
    return h1




def solve_mo1(nmrobj, mo_energy=None, mo_coeff=None, mo_occ=None,
              h1=None, s1=None):
    '''Solve the first order equation'''
    cput1 = (logger.process_clock(), logger.perf_counter())
    log = logger.Logger(nmrobj.stdout, nmrobj.verbose)
    if mo_energy is None:
        mo_energy = nmrobj._scf.mo_energy
    if mo_coeff is None:
        mo_coeff = nmrobj._scf.mo_coeff
    if mo_occ is None:
        mo_occ = nmrobj._scf.mo_occ

    mol = nmrobj.mol
    orboa = mo_coeff[0][:, mo_occ[0] > 0]
    orbob = mo_coeff[1][:, mo_occ[1] > 0]
    if h1 is None:
        dm0 = nmrobj._scf.make_rdm1(mo_coeff, mo_occ)
        h1 = nmrobj.get_fock(dm0)
        h1 = (lib.einsum('xpq,pi,qj->xij', h1[0], mo_coeff[0].conj(), orboa),
              lib.einsum('xpq,pi,qj->xij', h1[1], mo_coeff[1].conj(), orbob))
        cput1 = log.timer('first order Fock matrix', *cput1)
    if s1 is None:
        s1 = nmrobj.get_ovlp(mol)
        s1 = (lib.einsum('xpq,pi,qj->xij', s1, mo_coeff[0].conj(), orboa),
              lib.einsum('xpq,pi,qj->xij', s1, mo_coeff[1].conj(), orbob))

    # Always use CPHF
    vind = gen_vind(nmrobj._scf, mo_coeff, mo_occ)
    mo10, mo_e10 = ucphf.solve(vind, mo_energy, mo_occ, h1, s1,
                               nmrobj.max_cycle_cphf, nmrobj.conv_tol,
                               verbose=log)

    logger.timer(nmrobj, 'solving mo1 eqn', *cput1)
    return mo10, mo_e10


def gen_vind(mf, mo_coeff, mo_occ):
    '''Induced potential'''
    vresp = mf.gen_response(hermi=2)
    occidxa = mo_occ[0] > 0
    occidxb = mo_occ[1] > 0
    orboa = mo_coeff[0][:, occidxa]
    orbob = mo_coeff[1][:, occidxb]
    nocca = orboa.shape[1]
    noccb = orbob.shape[1]
    nao, nmo = mo_coeff[0].shape

    def vind(mo1):
        mo1a = mo1[:, :nocca*nmo].reshape(-1, nmo, nocca)
        mo1b = mo1[:, nocca*nmo:].reshape(-1, nmo, noccb)
        # Vectorized version: perform matrix multiplication in batch for both alpha and beta
        dm1a_intermediate = mo_coeff[0] @ mo1a @ orboa.T.conj()  # shape: (nbatch, nao, nao)
        dm1b_intermediate = mo_coeff[1] @ mo1b @ orbob.T.conj()  # shape: (nbatch, nao, nao)
        # Compute d1 - d1.conj().T for all matrices in batch
        dm1a = dm1a_intermediate - dm1a_intermediate.transpose(0, 2, 1).conj()
        dm1b = dm1b_intermediate - dm1b_intermediate.transpose(0, 2, 1).conj()
        dm1 = numpy.asarray((dm1a, dm1b))
        v1ao = vresp(dm1)
        # Vectorized version: perform matrix multiplication in batch for both alpha and beta
        v1a = mo_coeff[0].T.conj() @ v1ao[0] @ orboa  # shape: (nbatch, nmo_alpha, nocc_alpha)
        v1b = mo_coeff[1].T.conj() @ v1ao[1] @ orbob  # shape: (nbatch, nmo_beta, nocc_beta)
        v1mo = numpy.hstack((v1a.reshape(len(v1a), -1),
                             v1b.reshape(len(v1b), -1)))
        return v1mo.ravel()
    return vind


class NMR(rhf_nmr.NMR):

    def shielding(self):
        if getattr(self._scf, 'spin_square', None):
            s2 = self._scf.spin_square()[0]
            if s2 > 1e-4:
                logger.warn(self, '<S^2> = %s. UHF-NMR shielding may have large error.\n'
                            'paramagnetic NMR should include this result plus '
                            'g-tensor and HFC tensors.', s2)
        return rhf_nmr.NMR.shielding(self)

    dia = dia
    para = para
    get_fock = get_fock
    solve_mo1 = solve_mo1


scf.uhf.UHF.sNMR = lib.class_as_method(NMR)



# Copyright 2014-2022 The PySCF Developers, 2025-2026 Sidereus-AI ltd. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");

import numpy as np
from pyscf import lib, scf, hessian
from pyscf.hessian import thermo
from pyscf.data import nist
from pyscf.lib import logger
from ..utils.matplotlib import plot_spectrum
# from ..utils.scale import fundamentals as data

def polarizability_derivative_numerical_dEdE(mf, dE=2.5e-3):
    # Return in (   natm, 3,           3, 3      )
    #            < derivative > < polarizability >
    #
    # This function makes the mf object unusable, please make a new one after calling this function.
    mol = mf.mol

    with mol.with_common_orig((0, 0, 0)):
        dipole_integral = mol.intor('int1e_r').reshape(3, mol.nao, mol.nao)
        # int1e_irp: < | r p | >. Used for basis function derivatives approximation?
        # In PySCF, int1e_irp is (3, 3, nao, nao) flattened to (9, nao, nao).
        dipole_integral_derivative = -mol.intor('int1e_irp').reshape(3, 3, mol.nao, mol.nao)

    Hcore = mf.get_hcore()
    
    # Check for solvent (Note: GPU4PySCF specific solvent logic removed/simplified for generic PySCF)
    # If standard PySCF solvent is used (e.g. CPCM), gradients usually handle it if correctly initialized.
    # But here we are hacking the hcore to add E-field.
    # For now, we assume vacuum or that the solvent model is robust to get_hcore changes if it wraps it.
    
    # However, if we need to freeze solvent reaction field, that's complex without specific API.
    # We proceed with standard SCF gradient logic.

    def get_gradient_at_E(mf, E):
        # E is shape (3,)
        
        # Monkey patch get_hcore to include electric field
        original_get_hcore = mf.get_hcore
        def new_get_hcore(mol=None):
            # h = original_get_hcore(mol) # avoid recursion if original is bound?
            # Actually simpler: assume Hcore (cached) + field
            return Hcore + np.einsum('x,xij->ij', E, dipole_integral)
        
        mf.get_hcore = new_get_hcore
        
        try:
            mf.kernel()
            if not mf.converged:
                logger.warn(mf, "SCF not converged in Raman finite difference step with E=%s", E)
            
            dm = mf.make_rdm1()
            if dm.ndim == 3: # UHF/UKS
                dm = dm[0] + dm[1]
            
            # Compute Nuclear Gradient
            # For standard PySCF, we can just call the gradient method.
            # However, the gradient method might recalculate hcore using the *original* mol/logic if not careful,
            # but usually it calls mf.get_hcore().
            # BUT, the Hellmann-Feynman term from the added E-field perturbation must be added manually
            # because the standard grad method doesn't know about the E-field added in get_hcore 
            # (unless we used a proper CPHF/CPKS solver or supported external potential).
            
            grad_method = mf.nuc_grad_method()
            gradient = grad_method.kernel()
            
            # Add explicit derivative of the E-field interaction term: d/dR <psi| -E.mu |psi>
            # The term added to H is E.mu.
            # We need d/dR (E.mu) expectation value.
            # This involves basis function derivatives.
            
            aoslices = mol.aoslice_by_atom()
            for i_atom in range(mol.natm):
                p0, p1 = aoslices[i_atom][2:]

                d_dipoleintegral_dA = np.zeros((3, 3, mol.nao, mol.nao))
                # Logic from reference adapted to numpy
                d_dipoleintegral_dA[:, :, :, p0:p1] += dipole_integral_derivative[:, :, :, p0:p1]
                d_dipoleintegral_dA[:, :, p0:p1, :] += dipole_integral_derivative[:, :, :, p0:p1].transpose(0, 1, 3, 2)
                d_dipoleintegral_dA = d_dipoleintegral_dA.transpose(1,0,2,3) # (3_grad, 3_field, nao, nao)

                # Contribution: Tr(dm * d(E.mu)/dA) = Tr(dm * E_j * d(mu_j)/dA)
                # d_dipoleintegral_dA[x, y] is d(mu_y)/dA_x
                
                term = np.einsum('xyij,ij->xy', d_dipoleintegral_dA, dm) # (3_grad, 3_field)
                gradient[i_atom, :] += np.dot(term, E)

        finally:
            # Restore (though we iterate, so maybe not strictly necessary if we reset carefully, 
            # but good practice)
            mf.get_hcore = original_get_hcore

        return gradient

    dpdx = np.empty((mol.natm, 3, 3, 3))

    E_0 = np.zeros(3)
    gradient_0 = get_gradient_at_E(mf, E_0)

    for i_xyz in range(3):
        for j_xyz in range(i_xyz + 1, 3):
            E_pp = np.zeros(3)
            E_pp[i_xyz] += dE
            E_pp[j_xyz] += dE
            gradient_pp = get_gradient_at_E(mf, E_pp)

            E_pm = np.zeros(3)
            E_pm[i_xyz] += dE
            E_pm[j_xyz] -= dE
            gradient_pm = get_gradient_at_E(mf, E_pm)

            E_mp = np.zeros(3)
            E_mp[i_xyz] -= dE
            E_mp[j_xyz] += dE
            gradient_mp = get_gradient_at_E(mf, E_mp)

            E_mm = np.zeros(3)
            E_mm[i_xyz] -= dE
            E_mm[j_xyz] -= dE
            gradient_mm = get_gradient_at_E(mf, E_mm)

            dpdx_ij = (gradient_pp + gradient_mm - gradient_pm - gradient_mp) / (4 * dE**2)
            dpdx[:, :, i_xyz, j_xyz] = dpdx_ij
            dpdx[:, :, j_xyz, i_xyz] = dpdx_ij

        E_p = np.zeros(3)
        E_p[i_xyz] += dE
        gradient_p = get_gradient_at_E(mf, E_p)

        E_m = np.zeros(3)
        E_m[i_xyz] -= dE
        gradient_m = get_gradient_at_E(mf, E_m)

        dpdx[:, :, i_xyz, i_xyz] = (gradient_p + gradient_m - 2 * gradient_0) / (dE**2)
    
    # Reference implementation does `dpdx *= -1`.
    # Let's verify why.
    # Gradient is dE/dR.
    # We want d(alpha)/dR.
    # alpha = - d^2 E / dE^2.
    # So d(alpha)/dR = - d^3 E / dR dE^2 = - d(d^2 E/dR dE)/dE ...
    # Our finite diff calculates d^2 (Gradient) / dE^2 = d^2 (dE/dR) / dE^2 = d^3 E / dR dE^2.
    # So indeed, we need to negate it to get d(alpha)/dR.
    dpdx *= -1
    
    return dpdx

def eval_raman_intensity(mf, hessian=None):
    '''
    Main driver of Raman spectra intensity

    Args:
        mf: mean field object
        hessian: the hessian matrix in shape (natm, natm, 3, 3), if available

    Returns:
        node frequency: in cm^-1
        Raman scattering activity: in Angstrom**4 / AMU
        Depolarization ratio: dimensionless
    '''
    mol = mf.mol

    if hessian is None:
        hess_obj = mf.Hessian()
        hess_obj.auxbasis_response = 2 # specific to density fitting? standard hessian might not need this
        hessian = hess_obj.kernel()
    
    # Ensure hessian shape
    if hessian.ndim == 2:
        hessian = hessian.reshape(mol.natm, 3, mol.nao, 3).transpose(0, 2, 1, 3)
    
    assert hessian.shape == (mol.natm, mol.natm, 3, 3)

    freq_info = thermo.harmonic_analysis(mol, hessian)

    norm_mode = freq_info['norm_mode']
    freq_wavenumber = freq_info['freq_wavenumber']

    mf_copy = mf.copy() # Preserve the original mf
    dalpha_dR = polarizability_derivative_numerical_dEdE(mf_copy)
    dalpha_dQ = lib.einsum('AdEe,iAd->iEe', dalpha_dR, norm_mode)

    n_mode = len(freq_wavenumber)
    raman_intensities = np.zeros(n_mode)
    depolarization_ratio = np.zeros(n_mode)

    for i_mode in range(n_mode):
        dalpha_dQi = dalpha_dQ[i_mode]
        alpha_prime = 1.0/3.0 * (dalpha_dQi[0,0] + dalpha_dQi[1,1] + dalpha_dQi[2,2])
        alpha_prime_square = alpha_prime**2
        beta_prime_square = 0.5 * (
            + (dalpha_dQi[0,0] - dalpha_dQi[1,1])**2
            + (dalpha_dQi[0,0] - dalpha_dQi[2,2])**2
            + (dalpha_dQi[1,1] - dalpha_dQi[2,2])**2
            + 6 * (dalpha_dQi[0,1]**2 + dalpha_dQi[0,2]**2 + dalpha_dQi[1,2]**2)
        )

        raman_intensities[i_mode] = 45 * alpha_prime_square + 7 * beta_prime_square
        denominator = 45 * alpha_prime_square + 4 * beta_prime_square
        if abs(denominator) > 1e-12:
            depolarization_ratio[i_mode] = 3 * beta_prime_square / denominator
        else:
            depolarization_ratio[i_mode] = 0.0

    raman_intensities *= nist.BOHR**4

    return freq_wavenumber, raman_intensities, depolarization_ratio

class Raman(lib.StreamObject):
    def __init__(self, mf):
        self.mol = mf.mol
        self.verbose = mf.mol.verbose
        self.stdout = mf.mol.stdout
        self._scf = self.base = mf
        
        self.mf_hess = None
        self.vib_dict = None
        
        self.raman_intensities = NotImplemented
        self.depolarization_ratio = NotImplemented
        self.freq_wavenumber = NotImplemented

    def kernel(self, hessian=None):
        if self._scf.__module__.startswith("gpu4pyscf"):
            from gpu4pyscf.properties.raman import eval_raman_intensity as eval_raman_gpu
            freq, intensity, depolar = eval_raman_gpu(self._scf, hessian)
            self.freq_wavenumber = freq
            self.raman_intensities = intensity
            self.depolarization_ratio = depolar
        else:
            freq, intensity, depolar = eval_raman_intensity(self._scf, hessian)
            self.freq_wavenumber = freq
            self.raman_intensities = intensity
            self.depolarization_ratio = depolar
            
        self.vib_dict = {
            "freq_wavenumber": self.freq_wavenumber,
            "raman_intensities": self.raman_intensities,
            "depolarization_ratio": self.depolarization_ratio
        }
        return self.raman_intensities

    def summary(self, spectrum=None, w=50):
        log = logger.new_logger(self, 2)
        if self.raman_intensities is NotImplemented:
            self.kernel()
            
        log.log(
            "----------------------------------------------------------------------\n"
            " Mode      Frequency       Intensity       Depolarization Ratio       \n"
            "   #         cm^-1           A^4/AMU                                  \n"
            "----------------------------------------------------------------------")
            
        for i, (f, inten, depol) in enumerate(zip(self.freq_wavenumber, self.raman_intensities, self.depolarization_ratio)):
            flag_im = np.imag(f) > 1e-10
            chr_im = "i" if flag_im else " "
            log.log("{:5d}   {:12.4f}{:}   {:12.4f}     {:12.4f}".format(
                i, np.abs(f), chr_im, inten, depol))
        log.log("----------------------------------------------------------------------\n")
        
        if spectrum:
            # Similar plotting logic as IR but for Raman
            freq = self.freq_wavenumber
            inten = self.raman_intensities.copy()
            # Remove imaginary or negative freqs
            valid = (np.abs(np.imag(freq)) <= 1e-10) & (np.real(freq) > 0)
            freq = np.real(freq)
            
            # Mask
            inten[~valid] = 0
            
            xmax = max([4000, max(freq) + 5*max(w) if len(freq)>0 else 4000])
            
            fig, _, _ = plot_spectrum(freq, inten, w, 5*w, 4000, 0, xmax,
                                      "Vibration Wavenumber (cm$^{-1}$)",
                                      "Raman Activity (A$^4$/AMU)",
                                      "Raman Activity",
                                      "Raman Activity",
                                      "Raman Activity")
            fig.savefig(spectrum)
            fig.clf()

scf.hf.RHF.sRAMAN = lib.class_as_method(Raman)



# Copyright 2014-2022 The PySCF Developers, 2025-2026 Sidereus-AI ltd. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");


from pyscf import hessian, lib, scf
from pyscf.hessian import thermo
from pyscf.hessian.rhf import Hessian
from pyscf.lib import logger

import numpy as np
from scipy.constants import physical_constants

from ..utils.matplotlib import plot_spectrum
from ..utils.scale import fundamentals as data


def proc_hessian_(mf_hess, mo_energy=None, mo_coeff=None, mo_occ=None, h1ao_grad=None):
    """
    Process hessian evaluation for infrared analysis

    This function reuses hessian evaluation intermediate, mo_coeff rotation matrix (U matrix),
    to be further contracted with dipole tensor, without computing CP-HF equation a second time.
    """
    mf = mf_hess.base

    mo_energy = mo_energy if mo_energy else mf.mo_energy
    mo_coeff = mo_coeff if mo_coeff else mf.mo_coeff
    mo_occ = mo_occ if mo_occ else mf.mo_occ
    h1ao_grad = h1ao_grad if h1ao_grad else mf_hess.make_h1(mo_coeff, mo_occ)

    moao1_grad, mo_e1_grad = mf_hess.solve_mo1(
        mo_energy, mo_coeff, mo_occ, h1ao_grad)

    hess_elec = mf_hess.hess_elec(
        mo1=moao1_grad, mo_e1=mo_e1_grad, h1ao=h1ao_grad)
    hess_nuc = mf_hess.hess_nuc()
    mf_hess.de = hess_elec + hess_nuc
    mo1_grad = lib.einsum("up, uv, Axvi -> Axpi",
                          mo_coeff, mf.get_ovlp(), moao1_grad)
    return mf_hess, h1ao_grad, mo1_grad, mo_e1_grad


def get_h2ao_dipderiv(mf):
    """
    Generate core hamiltonian derivative of gradient-dipole

    Returns
    -------
    h2ao
        5-dim tensor: (atom number, atom coordinate components, dipole components, basis number, basis number)
    """
    mol = mf.mol
    natm, nao = mol.natm, mol.nao

    h2ao = np.zeros((natm, 3, 3, nao, nao))
    int1e_irp = mol.intor("int1e_irp").reshape(3, 3, nao, nao).swapaxes(0, 1)
    for a in range(natm):
        _, _, a1, a2 = mol.aoslice_by_atom()[a]
        h2ao[a, :, :, :, a1:a2] = int1e_irp[:, :, :, a1:a2]
    h2ao += h2ao.swapaxes(-1, -2)

    return h2ao


def kernel_dipderiv(mf_ir):
    """
    Dipole derivative main driver

    Dipole derivative is intermediate quantity for IR intensity evaluation.
    Note that if dipole derivative is the only purpose, then CP-HF equation wrt gradient rotation
    is not an efficient way, instead dipole rotation should make things easier. However, since
    hessian evaluation is inevitable when performing vibrational analysis and IR intensity,
    so we drop procedure that uses efficient CP-HF wrt dipole rotation.

    Returns
    -------
    de
        3-dim tensor: (atom number, atom coordinate components, dipole components)
    """
    if mf_ir.mf_hess is None:
        mf_ir.mf_hess = mf_ir.hess_cls(mf_ir.base)
        # mf_ir.mf_hess.grid_response = True  # in case mf_hess is dft method, but seems no effect?
    if mf_ir.mf_hess.de is None \
            or np.abs(mf_ir.mf_hess.de).sum() < 1e-14 \
            or mf_ir._mo1_grad is NotImplemented:
        mf_hess, h1ao_grad, mo1_grad, mo_e1_grad = proc_hessian_(mf_ir.mf_hess)
        mf_ir._h1ao_grad = h1ao_grad
        mf_ir._mo1_grad = mo1_grad
        mf_ir._mo_e1_grad = mo_e1_grad
    mo1_grad = mf_ir._mo1_grad

    if mf_ir._h1_dip is NotImplemented:
        mf = mf_ir.base
        mo_coeff = mf.mo_coeff
        mo_occ = mf.mo_occ
        orbo = mo_coeff[:, mo_occ > 0]
        int_r = mf.mol.intor_symmetric("int1e_r")
        h1_dip = lib.einsum("tuv, up, vi-> tpi", int_r, mo_coeff, orbo)
        mf_ir._h1_dip = h1_dip
    h1_dip = mf_ir._h1_dip

    if mf_ir._h2ao_dipderiv is NotImplemented:
        mf_ir._h2ao_dipderiv = get_h2ao_dipderiv(mf_ir.base)
    h2ao_dipderiv = mf_ir._h2ao_dipderiv

    mol = mf_ir.mol
    natm = mol.natm
    de = np.zeros((natm, 3, 3))
    for a in range(natm):
        de[a] = np.eye(3) * mol.atom_charge(a)
    de += np.einsum("Axtuv, uv -> Axt", h2ao_dipderiv, mf_ir.base.make_rdm1())
    de -= 4 * np.einsum("tpi, Axpi -> Axt", h1_dip, mo1_grad)

    mf_ir.de = de
    return de


def kernel_ir(mf_ir):
    """
    Infrared spectra intensity main driver

    Returns
    -------
    ir_inten
        1-dim vector; unit km/mol; same length to number of vibrational modes.

    See Also
    --------
    For unit conversion, also see
    https://github.com/psi4/psi4/blob/v1.5/psi4/driver/qcdb/vib.py#L589-L590
    """
    mol = mf_ir.mol
    if mf_ir.vib_dict is None:
        mf_ir.vib_dict = thermo.harmonic_analysis(mol, mf_ir.mf_hess.de)
    d = mf_ir.vib_dict
    q = d["norm_mode"].reshape(-1, mol.natm * 3)
    de = mf_ir.de.reshape(-1, 3)
    de_q = np.dot(q, de)

    alpha = physical_constants["fine-structure constant"][0]
    amu = physical_constants["atomic mass constant"][0]
    m_e = physical_constants["electron mass"][0]
    N_A = physical_constants["Avogadro constant"][0]
    a_0 = physical_constants["Bohr radius"][0]
    # unit_kmmol ~= 974.88011
    # another way is `(e_c * N_A)**2 / 10000000 * np.pi / 3`
    #     however, above formula uses deprecated definition of vacuum magnetic permeability value
    #     see also https://en.wikipedia.org/wiki/Vacuum_permeability
    unit_kmmol = alpha**2 * (1e-3 / amu) * m_e * N_A * np.pi * a_0 / 3

    ir_inten = unit_kmmol * np.einsum("qt, qt -> q", de_q, de_q)

    mf_ir.ir_inten = np.round(ir_inten, decimals=4)
    return ir_inten


class Infrared(lib.StreamObject):
    hess_cls = hessian.rhf.Hessian

    def __init__(self, mf):
        mol = mf.mol
        self.mol = mol
        self.verbose = mol.verbose
        self.stdout = mol.stdout
        self._scf = self.base = mf  # type: scf.hf.RHF

        self.mf_hess = None  # type: Hessian
        self.vib_dict = None  # type: dict

        self._h1ao_grad = NotImplemented
        self._mo_e1_grad = NotImplemented
        self._mo1_grad = NotImplemented
        self._h1_dip = NotImplemented
        self._h2ao_dipderiv = NotImplemented

        self.de = NotImplemented
        self.ir_inten = NotImplemented
        self.scaled_fundamental = NotImplemented

    kernel_dipderiv = kernel_dipderiv
    kernel_ir = kernel_ir

    def kernel(self, *args, **kwargs):
        if self._scf.__module__.startswith("gpu4pyscf"):
            from gpu4pyscf.properties.ir import eval_ir_freq_intensity
            mf = self._scf
            freq, intensity = eval_ir_freq_intensity(mf, mf.Hessian())
            self.ir_inten = np.round(intensity.get(), decimals=4)
            self.vib_dict = {"freq_wavenumber": freq.get()}
        else:
            self.kernel_dipderiv()
            self.kernel_ir()
        return self.ir_inten

    def summary(self, spectrum=None, I=None, F=None, s=None, w=50, ):
        log = logger.new_logger(self, 2)
        if I is None:
            I = self.ir_inten
        if F is None:
            F = self.vib_dict["freq_wavenumber"]
        if s is None:
            basis = self.mol.basis.lower().replace("-", "")
            xc = getattr(self._scf, "xc", "").lower().replace("-", "")
            key = f"{xc}/{basis}"
            if key in data:
                s = data[key]
            else:
                log.warn(f"红外基频因子 {key} 未找到，使用默认值 1.0")
                s = 1
        self.scaled_fundamental = F * s
        if isinstance(w, (int, float)):
            w = np.ones_like(F) * w
        log.log(
            "------------------------------------------\n"
            " Mode      Frequency       Intensity      \n"
            "   #         cm^-1          km/mol        \n"
            "------------------------------------------")

        for i, (f, inten) in enumerate(zip(self.scaled_fundamental, self.ir_inten)):
            flag_im = np.imag(f) > 1e-10
            chr_im = "i" if flag_im else " "
            log.log("{:5d}   {:12.4f}{:}   {:12.4f}".format(
                i, np.abs(f), chr_im, inten))
        log.log("------------------------------------------\n")
        if spectrum:
            freq = self.scaled_fundamental
            ir_inten = self.ir_inten.copy()
            ir_inten[np.abs(np.imag(freq)) > 1e-10] = 0
            ir_inten[freq < 0] = 0
            freq = np.real(freq)
            xmax = max([4000, max(freq) + 5*max(w)])
            fig, _, _ = plot_spectrum(freq, ir_inten, w, 5*w, 4000, 0, xmax,
                                      "Vibration Wavenumber (cm$^{-1}$)",
                                      "Molar Absorption Coefficient (L mol$^{-1}$ cm$^{-1}$)",
                                      "IR Intensity (km mol$^{-1}$)",
                                      "Molar Absorption Coefficient",
                                      "IR Intensity")
            fig.savefig(spectrum)
            fig.clf()


scf.hf.RHF.sIR = lib.class_as_method(Infrared)



from .rhf import OPT as rhf_OPT
from pyscf import dft, lib


class OPT(rhf_OPT):
    method = "dft"
    unrestricted = False


dft.rks.RKS.sOPT = dft.rks_symm.RKS.sOPT = lib.class_as_method(OPT)
try:
    from gpu4pyscf.dft import RKS
    RKS.sOPT = lib.class_as_method(OPT)
except ImportError:
    pass



from .rhf import OPT as rhf_OPT
from pyscf import scf, lib


class OPT(rhf_OPT):
    method = "scf"
    unrestricted = True


scf.uhf.UHF.sOPT = lib.class_as_method(OPT)
try:
    from gpu4pyscf.scf import UHF
    UHF.sOPT = lib.class_as_method(OPT)
except ImportError:
    pass



# Copyright 2014-2022 The PySCF Developers, 2025-2026 Sidereus-AI ltd. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");


from pyscf import dft
import numpy
from pyscf import lib
from pyscf.lib import logger
from pyscf.scf import _vhf
from pyscf.dft import numint
from . import rhf as rhf_nmr
from . import rks as rks_nmr
from . import uhf as uhf_nmr


def get_vxc_giao(ni, mol, grids, xc_code, dms, max_memory=2000, verbose=None):
    xctype = ni._xc_type(xc_code)
    make_rho, nset, nao = ni._gen_rho_evaluator(mol, dms, hermi=1)
    ngrids = len(grids.weights)
    BLKSIZE = numint.BLKSIZE
    blksize = int(max_memory*1e6/(12*nao*8*BLKSIZE))
    blksize = max(4, min(blksize, ngrids//BLKSIZE+1, 1200)) * BLKSIZE
    shls_slice = (0, mol.nbas)
    ao_loc = mol.ao_loc_nr()

    vmat = numpy.zeros((2, 3, nao, nao))
    if xctype == 'LDA':
        buf = numpy.empty((4, blksize, nao))
        ao_deriv = 0
        for ao, mask, weight, coords \
                in ni.block_loop(mol, grids, nao, ao_deriv, max_memory,
                                 blksize=blksize, buf=buf):
            rho_a = make_rho(0, ao, mask, 'LDA')
            rho_b = make_rho(1, ao, mask, 'LDA')
            vxc = ni.eval_xc(xc_code, (rho_a, rho_b), 1, deriv=1)[1]
            vrho = vxc[0]
            giao = mol.eval_gto('GTOval_ig', coords, comp=3,
                                non0tab=mask, out=buf[1:])
            aow = numpy.einsum('pi,p->pi', ao, weight*vrho[:, 0])
            vmat[0, 0] += numint._dot_ao_ao(mol, aow,
                                            giao[0], mask, shls_slice, ao_loc)
            vmat[0, 1] += numint._dot_ao_ao(mol, aow,
                                            giao[1], mask, shls_slice, ao_loc)
            vmat[0, 2] += numint._dot_ao_ao(mol, aow,
                                            giao[2], mask, shls_slice, ao_loc)
            aow = numpy.einsum('pi,p->pi', ao, weight*vrho[:, 1])
            vmat[1, 0] += numint._dot_ao_ao(mol, aow,
                                            giao[0], mask, shls_slice, ao_loc)
            vmat[1, 1] += numint._dot_ao_ao(mol, aow,
                                            giao[1], mask, shls_slice, ao_loc)
            vmat[1, 2] += numint._dot_ao_ao(mol, aow,
                                            giao[2], mask, shls_slice, ao_loc)
            vxc = vrho = aow = None

    elif xctype == 'GGA':
        buf = numpy.empty((10, blksize, nao))
        ao_deriv = 1
        for ao, mask, weight, coords \
                in ni.block_loop(mol, grids, nao, ao_deriv, max_memory,
                                 blksize=blksize, buf=buf):
            rho_a = make_rho(0, ao, mask, 'GGA')
            rho_b = make_rho(1, ao, mask, 'GGA')
            vxc = ni.eval_xc(xc_code, (rho_a, rho_b), 1, deriv=1)[1]
            vrho, vsigma = vxc[:2]
            giao = mol.eval_gto('GTOval_ig', coords, 3,
                                non0tab=mask, out=buf[4:])

            wva = numpy.empty_like(rho_a)
            wva[0] = weight * vrho[:, 0]
            wva[1:] = rho_a[1:] * (weight * vsigma[:, 0] * 2)  # sigma_uu
            wva[1:] += rho_b[1:] * (weight * vsigma[:, 1])      # sigma_ud
            wvb = numpy.empty_like(rho_b)
            wvb[0] = weight * vrho[:, 1]
            wvb[1:] = rho_b[1:] * (weight * vsigma[:, 2] * 2)  # sigma_dd
            wvb[1:] += rho_a[1:] * (weight * vsigma[:, 1])      # sigma_ud

            aow = numpy.einsum('npi,np->pi', ao[:4], wva)
            vmat[0, 0] += numint._dot_ao_ao(mol, aow,
                                            giao[0], mask, shls_slice, ao_loc)
            vmat[0, 1] += numint._dot_ao_ao(mol, aow,
                                            giao[1], mask, shls_slice, ao_loc)
            vmat[0, 2] += numint._dot_ao_ao(mol, aow,
                                            giao[2], mask, shls_slice, ao_loc)
            aow = numpy.einsum('npi,np->pi', ao[:4], wvb)
            vmat[1, 0] += numint._dot_ao_ao(mol, aow,
                                            giao[0], mask, shls_slice, ao_loc)
            vmat[1, 1] += numint._dot_ao_ao(mol, aow,
                                            giao[1], mask, shls_slice, ao_loc)
            vmat[1, 2] += numint._dot_ao_ao(mol, aow,
                                            giao[2], mask, shls_slice, ao_loc)

            giao = mol.eval_gto('GTOval_ipig', coords, 9,
                                non0tab=mask, out=buf[1:])
            rks_nmr._gga_sum_(vmat[0], mol, ao, giao,
                              wva, mask, shls_slice, ao_loc)
            rks_nmr._gga_sum_(vmat[1], mol, ao, giao,
                              wvb, mask, shls_slice, ao_loc)
            vxc = vrho = vsigma = aow = None
    elif xctype == 'MGGA':
        raise NotImplementedError('meta-GGA')

    return vmat - vmat.transpose(0, 1, 3, 2)


def get_fock(nmrobj, dm0=None):
    '''First order Fock matrix wrt external magnetic field'''
    if dm0 is None:
        dm0 = nmrobj._scf.make_rdm1()

    mol = nmrobj.mol

    log = logger.Logger(nmrobj.stdout, nmrobj.verbose)
    log.debug('First-order GIAO Fock matrix')

    mf = nmrobj._scf
    ni = mf._numint
    omega, alpha, hyb = ni.rsh_and_hybrid_coeff(mf.xc, spin=mol.spin)

    mem_now = lib.current_memory()[0]
    max_memory = max(2000, mf.max_memory*.9-mem_now)
    # Attach mo_coeff and mo_occ to improve get_vxc_giao efficiency
    dm0 = lib.tag_array(dm0, mo_coeff=mf.mo_coeff, mo_occ=mf.mo_occ)
    h1 = -get_vxc_giao(ni, mol, mf.grids, mf.xc, dm0,
                       max_memory=max_memory, verbose=nmrobj.verbose)

    intor = mol._add_suffix('int2e_ig1')
    if abs(hyb) > 1e-10:
        vj, vk = rhf_nmr.get_jk(mol, dm0)
        h1 += vj[0] + vj[1] - hyb * vk
        if abs(omega) > 1e-10:
            with mol.with_range_coulomb(omega):
                h1 -= (alpha-hyb) * rhf_nmr.get_jk(mol, dm0)[1]
    else:
        vj = _vhf.direct_mapdm(intor, 'a4ij', 'lk->s1ij',
                               dm0, 3, mol._atm, mol._bas, mol._env)
        h1 -= vj[0] + vj[1]

    h1 -= .5 * mol.intor('int1e_giao_irjxp', 3)
    h1 -= mol.intor_asymmetric('int1e_ignuc', 3)
    if mol.has_ecp():
        h1 -= mol.intor_asymmetric('ECPscalar_ignuc', 3)
    h1 -= mol.intor('int1e_igkin', 3)

    if nmrobj.chkfile:
        lib.chkfile.dump(nmrobj.chkfile, 'nmr/h1', h1)
    return h1


def solve_mo1(nmrobj, mo_energy=None, mo_coeff=None, mo_occ=None,
              h1=None, s1=None):
    # Only use CPHF for hybrid functionals - but now always use CPHF
    # Call the parent function (which always uses CPHF)
    return uhf_nmr.solve_mo1(nmrobj, mo_energy, mo_coeff, mo_occ,
                             h1, s1)


class NMR(uhf_nmr.NMR):
    get_fock = get_fock
    solve_mo1 = solve_mo1


dft.uks.UKS.sNMR = dft.uks_symm.UKS.sNMR = lib.class_as_method(NMR)



from rdkit import Chem
from rdkit.Chem import AllChem, rdDetermineBonds
from rdkit.Chem.Draw import rdMolDraw2D


def pyscf2rdmol(pyscf_mol):
    mol_xyz = pyscf_mol.tostring(format="xyz")
    raw_mol = Chem.MolFromXYZBlock(mol_xyz)
    rd_mol = Chem.Mol(raw_mol)
    rdDetermineBonds.DetermineBonds(rd_mol, charge=pyscf_mol.charge)
    return rd_mol


def draw_rdmol_with_label(rd_mol, labels, filename, to2D=True):
    mol_to_draw = Chem.Mol(rd_mol)
    for atom_id, label in labels.items():
        atom = mol_to_draw.GetAtomWithIdx(atom_id)
        atom.SetProp("atomNote", label)
    if to2D:
        mol_to_draw.RemoveAllConformers()
        AllChem.Compute2DCoords(mol_to_draw)
    drawer = rdMolDraw2D.MolDraw2DCairo(600, 600)
    drawer.DrawMolecule(mol_to_draw)
    drawer.FinishDrawing()
    drawer.WriteDrawingText(filename)



# Copyright 2014-2022 The PySCF Developers, 2025-2026 Sidereus-AI ltd. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");


from .uhf import Infrared
from pyscf import hessian, dft, lib


class Infrared(Infrared):
    hess_cls = hessian.uks.Hessian

dft.uks.UKS.sIR = dft.uks_symm.UKS.sIR = lib.class_as_method(Infrared)



import subprocess
import tempfile
import sys
import yaml
from copy import deepcopy
from pathlib import Path
import numpy as np
from pyscf import lib, scf
import h5py

optimization_data = {
    "geom": {
        "type": "redund",
        "fn": None,
        "coord_kwargs": {},
    },
    "calc": {
        "type": "pyscf",
    },
}

growing_string_data = {
    "precontr": {},
    "cos": {
        "type": "gs",
    },
    "opt": {
        "type": "string",
        "stop_in_when_full": -1,
    },
    "calc": {
        "type": "pyscf",
    },
    "tsopt": {
        "type": "rsprfo",
    },
    "geom": {
        "type": "dlc",
        "fn": [None, None],
    },
}


class OPT(lib.StreamObject):
    method = "scf"
    unrestricted = False
    threshold_map = {
        "loose": "gau_loose",
        "tight": "gau_tight",
        "verytight": "gau_vtight",
    }

    def __init__(self, mf, ts=False, product_xyzs=None, threshold=None, max_cycles=None,
                 verbose=None, constraints=None, reaction_modes=None):
        self.mf = mf
        self.ts = ts
        self.product_xyzs = product_xyzs
        if threshold and threshold.lower() not in ("loose", "tight", "verytight", None):
            raise ValueError(f"Threshold should be one of \
loose, tight, verytight, None, but got {threshold}")
        self.threshold = threshold
        if constraints is not None:
            if not isinstance(constraints, list):
                raise TypeError("constraints should be a list of lists")
        self.constraints = constraints
        if reaction_modes is not None:
            if not isinstance(reaction_modes, list):
                raise TypeError(
                    "reaction_modes should be a list of primitive coordinate lists")
        self.reaction_modes = reaction_modes
        self.max_cycles = max_cycles
        self.verbose = verbose
        if self.verbose is None:
            self.verbose = mf.verbose

    def _gen_geom_block(self, optimization_data, mol):
        data = deepcopy(optimization_data)
        data["geom"]["fn"] = mol.tostring(format="xyz")
        if self.constraints is not None:
            data["geom"]["coord_kwargs"]["constrain_prims"] = self.constraints
        return data

    def _gen_ts_search_block(self, growing_string_data, mol):
        data = deepcopy(growing_string_data)
        data["geom"]["fn"] = [mol.tostring(format="xyz"), self.product_xyzs]
        return data

    def _gen_calc_block(self, data, mol):
        data["calc"]["method"] = self.method
        data["calc"]["unrestricted"] = self.unrestricted
        data["calc"]["basis"] = mol.basis
        data["calc"]["charge"] = mol.charge
        data["calc"]["mult"] = mol.spin + 1
        data["calc"]["pal"] = lib.num_threads()
        if getattr(self.mf, "xc", None):
            data["calc"]["xc"] = self.mf.xc
        if self.mf.__module__.startswith("gpu4pyscf"):
            data["calc"]["use_gpu"] = True
        if getattr(self.mf, "auxbasis", None):
            data["calc"]["auxbasis"] = self.mf.auxbasis

    def _gen_opt_block(self):
        opt_block = {"thresh": self.threshold_map.get(self.threshold, "gau")}
        if self.max_cycles is not None:
            opt_block["max_cycles"] = self.max_cycles
        if self.reaction_modes is not None:
            if not self.ts:
                raise KeyError(
                    "reaction_modes can only be used with TS optimization")
            opt_block["rx_modes"] = self.reaction_modes
        return opt_block

    def _run_yaml(self, data, logger, output_process):
        with tempfile.TemporaryDirectory() as tmpdir:
            tmpdir = Path(tmpdir)
            with open(tmpdir / "run.yaml", "w") as f:
                yaml.dump(data, f)
            process = subprocess.Popen(
                ["pysis", "run.yaml"],
                cwd=tmpdir,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                text=True,
                bufsize=1
            )
            for line in process.stdout:
                line = line.strip()
                if line and self.verbose >= lib.logger.NOTE:
                    logger.note("%s", line)
            stdout, stderr = process.communicate()
            if process.returncode != 0:
                raise RuntimeError(f"Optimization failed")
            output_process(tmpdir)

    def _opt_output_process(self, tmpdir):
        if self.ts:
            outname = "ts_opt.xyz"
        else:
            outname = "final_geometry.xyz"
        with open(tmpdir / outname, "r") as f:
            data = np.array([line.strip().split()[1:]
                            for line in f.readlines()[2:]], dtype=float)
        self.mf.mol.set_geom_(data)

    def kernel(self):
        if self.ts and self.product_xyzs:
            return self.ts_search()
        return self.optimize()

    def optimize(self):
        mol = self.mf.mol
        logger = lib.logger.new_logger(mol, self.verbose)
        # GEOM block
        data = self._gen_geom_block(optimization_data, mol)
        # CALC block
        self._gen_calc_block(data, mol)
        # OPT and TSOPT block
        opt_block = self._gen_opt_block()
        if not self.ts:
            data["opt"] = {"type": "rfo", **opt_block}
        else:
            data["tsopt"] = {"type": "rsprfo", **opt_block}

        self._run_yaml(data, logger, self._opt_output_process)

        return mol

    def ts_search(self):
        mol = self.mf.mol
        logger = lib.logger.new_logger(mol, self.verbose)
        data = self._gen_ts_search_block(growing_string_data, mol)
        self._gen_calc_block(data, mol)
        self._run_yaml(data, logger, self._opt_output_process)
        return mol

    def _irc_output_process(self, tmpdir):
        self.irc_results = {"first": None, "last": None}
        with open(tmpdir / "finished_first.xyz", "r") as f:
            self.irc_results["first"] = f.read()
        with open(tmpdir / "finished_last.xyz", "r") as f:
            self.irc_results["last"] = f.read()
        with h5py.File(tmpdir / "finished_irc_data.h5", "r") as f:
            self.irc_results["energies"] = f["energies"][:]
            self.irc_results["coordinates"] = f["coords"][:]

    def irc(self):
        mol = self.mf.mol
        logger = lib.logger.new_logger(mol, self.verbose)
        # GEOM block
        data = self._gen_geom_block(optimization_data, mol)
        # CALC block
        self._gen_calc_block(data, mol)
        # IRC block
        data["irc"] = {"type": "eulerpc"}
        self._run_yaml(data, logger, self._irc_output_process)
        return self.irc_results


scf.hf.RHF.sOPT = lib.class_as_method(OPT)
try:
    from gpu4pyscf.scf import RHF
    RHF.sOPT = lib.class_as_method(OPT)
except ImportError:
    pass



import pyscf
import pyscf.sidereus


class TestIR:
    def setup_method(self, method):
        self.atom = 'H 0 0 0; F 0 0 0.917'
        self.basis = 'sto-3g'
        self.xc = "LDA"

    def test_rhf(self):
        mol = pyscf.M(atom=self.atom, basis=self.basis)
        mol.RHF().run().sIR().kernel()

    def test_uhf(self):
        mol = pyscf.M(atom=self.atom, basis=self.basis)
        mol.UHF().run().sIR().kernel()

    def test_rks(self):
        mol = pyscf.M(atom=self.atom, basis=self.basis)
        mol.RKS(xc=self.xc).run().sIR().kernel()

    def test_uks(self):
        mol = pyscf.M(atom=self.atom, basis=self.basis)
        mol.UKS(xc=self.xc).run().sIR().kernel()



# Copyright 2014-2022 The PySCF Developers, 2025-2026 Sidereus-AI ltd. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");

from .rhf import Raman
from pyscf import lib, dft

class Raman(Raman):
    pass

dft.uks.UKS.sRAMAN = dft.uks_symm.UKS.sRAMAN = lib.class_as_method(Raman)
try:
    from gpu4pyscf.dft import UKS
    UKS.sRAMAN = lib.class_as_method(Raman)
except ImportError:
    pass



from .rhf import OPT as rhf_OPT
from pyscf import dft, lib


class OPT(rhf_OPT):
    method = "dft"
    unrestricted = True


dft.uks.UKS.sOPT = dft.uks_symm.UKS.sOPT = lib.class_as_method(OPT)
try:
    from gpu4pyscf.dft import UKS
    UKS.sOPT = lib.class_as_method(OPT)
except ImportError:
    pass



# Copyright 2014-2022 The PySCF Developers, 2025-2026 Sidereus-AI ltd. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");


from pyscf import scf
import numpy
from pyscf import lib
from pyscf.lib import logger
from pyscf.scf import _vhf
from pyscf.scf import cphf
from pyscf.scf import _response_functions  # noqa
from pyscf.data import nist
from ..utils.rdkit import pyscf2rdmol, draw_rdmol_with_label, Chem
from ..utils.matplotlib import plot_spectrum


def dia(nmrobj, shielding_nuc=None, dm0=None):
    '''Diamagnetic part of NMR shielding tensors.

    See also J. Olsen et al., Theor. Chem. Acc., 90, 421 (1995)
    '''
    if shielding_nuc is None:
        shielding_nuc = nmrobj.shielding_nuc
    if dm0 is None:
        dm0 = nmrobj._scf.make_rdm1()

    mol = nmrobj.mol
    mf = nmrobj._scf

    if getattr(mf, 'with_x2c', None):
        raise NotImplementedError('X2C for NMR shielding')

    if getattr(mf, 'with_qmmm', None):
        raise NotImplementedError('NMR shielding with QM/MM')

    if getattr(mf, 'with_solvent', None):
        raise NotImplementedError('NMR shielding with Solvent')

    msc_dia = []
    for n, atm_id in enumerate(shielding_nuc):
        with mol.with_rinv_origin(mol.atom_coord(atm_id)):
            # a11part = (B dot) -1/2 frac{\vec{r}_N}{r_N^3} r (dot mu)
            h11 = mol.intor('int1e_giao_a11part', comp=9)
            e11 = numpy.einsum('xij,ij->x', h11, dm0).reshape(3, 3)
            e11 = e11 - numpy.eye(3) * e11.trace()
            h11 = mol.intor('int1e_a01gp', comp=9)
            e11 += numpy.einsum('xij,ij->x', h11, dm0).reshape(3, 3)
        msc_dia.append(e11)
    return numpy.array(msc_dia).reshape(-1, 3, 3)


def para(nmrobj, mo10=None, mo_coeff=None, mo_occ=None, shielding_nuc=None):
    '''Paramagnetic part of NMR shielding tensors.
    '''
    if mo_coeff is None:
        mo_coeff = nmrobj._scf.mo_coeff
    if mo_occ is None:
        mo_occ = nmrobj._scf.mo_occ
    if shielding_nuc is None:
        shielding_nuc = nmrobj.shielding_nuc
    if mo10 is None:
        mo10 = nmrobj.solve_mo1()[0]

    mol = nmrobj.mol
    para_vir = numpy.empty((len(shielding_nuc), 3, 3))
    para_occ = numpy.empty((len(shielding_nuc), 3, 3))
    occidx = mo_occ > 0
    viridx = mo_occ == 0
    orbo = mo_coeff[:, occidx]
    orbv = mo_coeff[:, viridx]
    # *2 for double occupancy
    # Vectorized version: extract occupied and virtual parts and perform matrix multiplication in batch
    mo10_occ = mo10[:, occidx, :]  # shape: (3, nocc, nocc)
    mo10_vir = mo10[:, viridx, :]  # shape: (3, nvir, nocc)
    dm10_oo = orbo @ (mo10_occ*2) @ orbo.T.conj()  # shape: (3, nao, nao)
    dm10_vo = orbv @ (mo10_vir*2) @ orbo.T.conj()  # shape: (3, nao, nao)
    for n, atm_id in enumerate(shielding_nuc):
        mol.set_rinv_origin(mol.atom_coord(atm_id))
        # H^{01} = 1/2(A01 dot p + p dot A01) => (a01p + c.c.)/2 ~ <a01p>
        # Im[A01 dot p] = Im[vec{r}/r^3 x vec{p}] = Im[-i p (1/r) x p] = -p (1/r) x p
        h01i = mol.intor_asymmetric('int1e_prinvxp', 3)  # = -Im[H^{01}]
        # <H^{01},MO^1> = - Tr(Im[H^{01}],Im[MO^1]) = Tr(-Im[H^{01}],Im[MO^1])
        # *2 for + c.c.
        para_occ[n] = numpy.einsum('xji,yij->xy', dm10_oo, h01i) * 2
        # *2 for + c.c.
        para_vir[n] = numpy.einsum('xji,yij->xy', dm10_vo, h01i) * 2
    msc_para = para_occ + para_vir
    return msc_para, para_vir, para_occ


def make_h10(mol, dm0, verbose=logger.WARN):
    '''Imaginary part of first order Fock operator

    Note the side effects of set_common_origin
    '''
    log = logger.new_logger(mol, verbose)
    # A10_i dot p + p dot A10_i consistents with <p^2 g>
    # A10_j dot p + p dot A10_j consistents with <g p^2>
    # 1/2(A10_j dot p + p dot A10_j) => Im[1/4 (rjxp - pxrj)] = -1/2 <irjxp>
    log.debug('First-order GIAO Fock matrix')
    h1 = -.5 * mol.intor('int1e_giao_irjxp', 3) + make_h10giao(mol, dm0)
    return h1


def get_jk(mol, dm0):
    # J = Im[(i i|\mu g\nu) + (i gi|\mu \nu)] = -i (i i|\mu g\nu)
    # K = Im[(\mu gi|i \nu) + (\mu i|i g\nu)]
    #   = [-i (\mu g i|i \nu)] - h.c.   (-h.c. for anti-symm because of the factor -i)
    intor = mol._add_suffix('int2e_ig1')
    vj, vk = _vhf.direct_mapdm(intor,  # (g i,j|k,l)
                               'a4ij', ('lk->s1ij', 'jk->s1il'),
                               dm0, 3,  # xyz, 3 components
                               mol._atm, mol._bas, mol._env)
    vk = vk - numpy.swapaxes(vk, -1, -2)
    return -vj, -vk


def make_h10giao(mol, dm0):
    vj, vk = get_jk(mol, dm0)
    h1 = vj - .5 * vk
    # Im[<g\mu|H|g\nu>] = -i * (gnuc + gkin)
    h1 -= mol.intor_asymmetric('int1e_ignuc', 3)
    if mol.has_ecp():
        h1 -= mol.intor_asymmetric('ECPscalar_ignuc', 3)
    h1 -= mol.intor('int1e_igkin', 3)
    return h1


def make_s10(mol):
    '''First order overlap matrix wrt external magnetic field.'''
    # Im[<g\mu |g\nu>]
    s1 = -mol.intor_asymmetric('int1e_igovlp', 3)
    return s1


get_ovlp = make_s10


def solve_mo1(nmrobj, mo_energy=None, mo_coeff=None, mo_occ=None,
              h1=None, s1=None):
    '''Solve the first order equation'''
    if mo_energy is None:
        mo_energy = nmrobj._scf.mo_energy
    if mo_coeff is None:
        mo_coeff = nmrobj._scf.mo_coeff
    if mo_occ is None:
        mo_occ = nmrobj._scf.mo_occ

    cput1 = (logger.process_clock(), logger.perf_counter())
    log = logger.Logger(nmrobj.stdout, nmrobj.verbose)

    mol = nmrobj.mol
    orbo = mo_coeff[:, mo_occ > 0]
    if h1 is None:
        dm0 = nmrobj._scf.make_rdm1(mo_coeff, mo_occ)
        h1 = lib.einsum('xpq,pi,qj->xij', nmrobj.get_fock(dm0),
                        mo_coeff.conj(), orbo)
        cput1 = log.timer('first order Fock matrix', *cput1)
    if s1 is None:
        s1 = lib.einsum('xpq,pi,qj->xij', nmrobj.get_ovlp(mol),
                        mo_coeff.conj(), orbo)

    # Always use CPHF
    vind = gen_vind(nmrobj._scf, mo_coeff, mo_occ)
    mo10, mo_e10 = cphf.solve(vind, mo_energy, mo_occ, h1, s1,
                              nmrobj.max_cycle_cphf, nmrobj.conv_tol,
                              verbose=log)

    log.timer('solving mo1 eqn', *cput1)
    return mo10, mo_e10


def get_fock(nmrobj, dm0=None):
    r'''First order partial derivatives of Fock matrix wrt external magnetic
    field.  \frac{\partial F}{\partial B}
    '''
    if dm0 is None:
        dm0 = nmrobj._scf.make_rdm1()

    log = logger.Logger(nmrobj.stdout, nmrobj.verbose)
    h1 = make_h10(nmrobj.mol, dm0, log)
    if nmrobj.chkfile:
        lib.chkfile.dump(nmrobj.chkfile, 'nmr/h1', h1)
    return h1


def gen_vind(mf, mo_coeff, mo_occ):
    '''Induced potential'''
    vresp = mf.gen_response(singlet=True, hermi=2)
    occidx = mo_occ > 0
    orbo = mo_coeff[:, occidx]
    nocc = orbo.shape[1]
    nao, nmo = mo_coeff.shape

    def vind(mo1):
        # Vectorized version: perform matrix multiplication in batch
        mo1_batch = mo1.reshape(-1, nmo, nocc)
        # shape: (nbatch, nao, nao)
        dm1_intermediate = mo_coeff @ (mo1_batch*2) @ orbo.T.conj()
        # Compute d1 - d1.conj().T for all matrices in batch
        dm1 = dm1_intermediate - dm1_intermediate.transpose(0, 2, 1).conj()
        v1mo = lib.einsum('xpq,pi,qj->xij', vresp(dm1), mo_coeff.conj(), orbo)
        return v1mo.ravel()
    return vind


class NMR(lib.StreamObject):
    def __init__(self, scf_method, element="H"):
        self.mol = scf_method.mol
        self.verbose = scf_method.mol.verbose
        self.stdout = scf_method.mol.stdout
        self.chkfile = scf_method.chkfile
        self._scf = scf_method

        self.element = element
        if element is None:
            self.shielding_nuc = range(self.mol.natm)
        else:
            self.shielding_nuc = [i for i in range(self.mol.natm)
                                  if self.mol.atom_symbol(i) == element]
        self.max_cycle_cphf = 20
        self.conv_tol = 1e-9

        self.B = NotImplemented

        self._keys = set(self.__dict__.keys())

    def dump_flags(self, verbose=None):
        log = logger.new_logger(self, verbose)
        log.info('\n')
        log.info('******** %s for %s ********',
                 self.__class__, self._scf.__class__)
        log.info('gauge = GIAO')
        log.info('shielding for atoms %s', str(self.shielding_nuc))
        log.info('Solving MO10 eq with CPHF.')
        log.info('CPHF conv_tol = %g', self.conv_tol)
        log.info('CPHF max_cycle_cphf = %d', self.max_cycle_cphf)
        if not self._scf.converged:
            log.warn('Ground state SCF is not converged')
        return self

    # Note mo10 is the imaginary part of MO^1
    def kernel(self):
        if self._scf.__module__.startswith("gpu4pyscf"):
            self.gpu_shielding()
        else:
            self.shielding()
        return self.B

    def summary(self, B=None, w=None, avg_equiv=True, img=None, spectrum=None):
        # http://cheshirenmr.info/ScalingFactors.htm
        # from Lodewyk, M. W.; Siebert, M. R.; Tantillo, D. J. Chem. Rev. 2012, 112, 1839-1862.
        # Only for b3lyp/6-31G* and SMD(Chloroform) or gas
        if B is None:
            B = self.B
        with_solvent = getattr(self._scf, 'with_solvent', None)
        slopes = numpy.empty(len(self.shielding_nuc))
        intercepts = numpy.empty(len(self.shielding_nuc))
        for n, atom_id in enumerate(self.shielding_nuc):
            element = self.mol.atom_symbol(atom_id)
            if element == "H":
                slopes[n] = 1.0 / 0.9957
                intercepts[n] = 32.2884
            elif element == "C":
                slopes[n] = 1.0 / 0.9269
                intercepts[n] = 187.4743
            else:
                raise NotImplementedError(f"Scaling for {element}")
        b_iso = numpy.trace(B, axis1=-2, axis2=-1) / 3
        shifts = (intercepts - b_iso) * slopes
        if avg_equiv or img:
            rd_mol = pyscf2rdmol(self.mol)
        if avg_equiv:
            ranks = list(Chem.CanonicalRankAtoms(rd_mol, breakTies=False))
            new_shifts = numpy.copy(shifts)
            for i in range(len(self.shielding_nuc)):
                curr_atom_idx = self.shielding_nuc[i]
                curr_rank = ranks[curr_atom_idx]
                equiv_indices = [j for j, idx in enumerate(self.shielding_nuc)
                                 if ranks[idx] == curr_rank]
                if len(equiv_indices) > 1:
                    avg_val = numpy.mean(shifts[equiv_indices])
                    new_shifts[i] = avg_val
            shifts = new_shifts
        if img:
            labels = {
                j: f"{shifts[i]:.2f}"
                for i, j in enumerate(self.shielding_nuc)
            }
            draw_rdmol_with_label(rd_mol, labels, img)
        if spectrum:
            if self.element == "H":
                default_min = 0
                default_max = 13
                default_pad = 1
                default_lw = 0.1
            elif self.element == "C":
                default_min = 0
                default_max = 210
                default_pad = 10
                default_lw = 2.0
            x_min = min(min(shifts), default_min) - default_pad
            x_max = max(max(shifts), default_max) + default_pad
            lw = default_lw if w is None else w
            fig, ax, ax2 = plot_spectrum(shifts,
                                         width=lw,
                                         x_min=x_min,
                                         x_max=x_max,
                                         x_label='Chemical Shift (ppm)',
                                         y_label='Intensity',
                                         y2_label="Occupancy",
                                         y_legend="Simulated Spectrum",
                                         y2_legend="Stick Spectrum")
            ticks = ax2.get_yticks()
            ax2.set_yticks([t for t in ticks if t >= 0 and t == int(t)])
            ax.invert_xaxis()
            ax.set_yticklabels([])
            fig.savefig(spectrum)
            fig.clf()
        log = logger.new_logger(self, 2)
        log.log(
            "---------------------------------\n"
            " atom      Chemical Shift        \n"
            "   #         ppm                 \n"
            "---------------------------------")
        for i, atm_id in enumerate(self.shielding_nuc):
            log.log("%5d   %12.4f             " % (i, shifts[i]))
        log.log("---------------------------------\n")
        return shifts

    def shielding(self):
        cput0 = (logger.process_clock(), logger.perf_counter())
        self.check_sanity()
        self.dump_flags()

        unit_ppm = nist.ALPHA**2 * 1e6
        msc_dia = self.dia()

        mo10, mo_e10 = self.solve_mo1()
        msc_para, para_vir, para_occ = self.para(mo10=mo10)

        msc_dia *= unit_ppm
        msc_para *= unit_ppm
        para_vir *= unit_ppm
        para_occ *= unit_ppm
        msc_dia = numpy.round(msc_dia, decimals=4)
        msc_para = numpy.round(msc_para, decimals=4)
        e11 = numpy.round(msc_para + msc_dia, decimals=4)

        logger.timer(self, 'NMR shielding', *cput0)
        if self.verbose >= logger.NOTE:
            for i, atm_id in enumerate(self.shielding_nuc):
                _write(self.stdout, e11[i],
                       '\ntotal shielding of atom %d %s'
                       % (atm_id, self.mol.atom_symbol(atm_id)))
                _write(self.stdout, msc_dia[i], 'dia-magnetic contribution')
                _write(self.stdout, msc_para[i], 'para-magnetic contribution')
                if self.verbose >= logger.INFO:
                    _write(self.stdout, para_occ[i],
                           'occ part of para-magnetism')
                    _write(self.stdout, para_vir[i],
                           'vir part of para-magnetism')
        self.B = e11
        return e11

    def gpu_shielding(self):
        from gpu4pyscf.properties.shielding import eval_shielding
        msc_dia, msc_para = eval_shielding(self._scf)
        msc_para = numpy.round(msc_para[self.shielding_nuc], decimals=4)
        msc_dia = numpy.round(msc_dia[self.shielding_nuc], decimals=4)
        e11 = numpy.round(msc_para + msc_dia, decimals=4)
        self.B = e11.get()
        if self.verbose >= logger.NOTE:
            for i, atm_id in enumerate(self.shielding_nuc):
                _write(self.stdout, e11[i],
                       '\ntotal shielding of atom %d %s'
                       % (atm_id, self.mol.atom_symbol(atm_id)))
                _write(self.stdout, msc_dia[i],
                       'dia-magnetic contribution')
                _write(self.stdout, msc_para[i],
                       'para-magnetic contribution')

    dia = dia
    para = para
    get_fock = get_fock
    solve_mo1 = solve_mo1

    def get_ovlp(self, mol=None):
        if mol is None:
            mol = self.mol
        return get_ovlp(mol)


def _write(stdout, msc3x3, title):
    stdout.write('%s\n' % title)
    stdout.write('B_x %s\n' % str(msc3x3[0]))
    stdout.write('B_y %s\n' % str(msc3x3[1]))
    stdout.write('B_z %s\n' % str(msc3x3[2]))
    stdout.flush()


scf.hf.RHF.sNMR = lib.class_as_method(NMR)



# Copyright 2014-2022 The PySCF Developers, 2025-2026 Sidereus-AI ltd. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");


from .rhf import Infrared
from pyscf import hessian, dft, lib


class Infrared(Infrared):
    hess_cls = hessian.rks.Hessian


dft.rks.RKS.sIR = dft.rks_symm.RKS.sIR = lib.class_as_method(Infrared)
try:
    from gpu4pyscf.dft import RKS
    RKS.sIR = lib.class_as_method(Infrared)
except ImportError:
    pass



import numpy as np
import matplotlib.pyplot as plt


def plot_spectrum(shifts, peak=1, width=2, padding=0, res=1000,
                  x_min=None, x_max=None, x_label=None,
                  y_label=None, y2_label=None, y_legend=None, y2_legend=None):
    if x_min is None:
        x_min = min(shifts) - padding
    if x_max is None:
        x_max = max(shifts) + padding
    if isinstance(peak, (int, float)):
        peak = np.ones_like(shifts) * peak
    if isinstance(width, (int, float)):
        width = np.ones_like(shifts) * width
    fig, ax = plt.subplots(figsize=(10, 6))
    x = np.linspace(x_min, x_max, res)
    y = np.zeros_like(x)
    y2 = np.zeros_like(x)
    gamma = width / 2.0
    for s, p, w in zip(shifts, peak, width):
        g = w / 2.0
        y += p * (1.0 / np.pi) * (g / ((x - s)**2 + g**2))
    ax.plot(x, y, color='black', lw=1.5, label=y_legend)
    if x_label is not None:
        ax.set_xlabel(x_label)
    if y_label is not None:
        ax.set_ylabel(y_label)
    if y_legend is not None:
        ax.legend(loc="upper left")
    ax2 = ax.twinx()
    if y2_label is not None:
        ax2.set_ylabel(y2_label)
    bin_width = 0.01
    bins = np.arange(x_min, x_max + bin_width, bin_width)
    counts, bin_edges = np.histogram(shifts, bins=bins, weights=peak)
    bin_centers = (bin_edges[:-1] + bin_edges[1:]) / 2
    mask = counts > 0
    markerline, stemlines, baseline = ax2.stem(bin_centers[mask], counts[mask],
                                               linefmt='C2-', markerfmt=' ',
                                               basefmt=' ', label=y2_legend)
    if y2_legend is not None:
        ax2.legend(loc="upper right")
    fig.tight_layout()
    ax.grid()
    return fig, ax, ax2



#  "Computational Thermochemistry: Scale Factor Databases and Scale Factors for Vibrational Frequencies Obtained from Electronic Model Chemistries," I. M. Alecu, J. Zheng, Y. Zhao, and D. G. Truhlar, J. Chem. Theory Comput. 6, 2872-2887 (2010).
fundamentals = {
    "b3lyp/def2tzvp": 0.959,
}



import pyscf
import pyscf.sidereus


class TestMINOPT:
    def setup_method(self, method):
        self.atom = "H 0 0 0; H 0 0 1"
        self.basis = "sto-3g"
        self.xc = "LDA"

    def test_rhf(self):
        mol = pyscf.M(atom=self.atom, basis=self.basis)
        mol.RHF().sOPT().kernel()

    def test_rks(self):
        mol = pyscf.M(atom=self.atom, basis=self.basis)
        mol.RKS(xc=self.xc).sOPT().kernel()

    def test_uhf(self):
        mol = pyscf.M(atom=self.atom, basis=self.basis)
        mol.UHF().sOPT().kernel()

    def test_uks(self):
        mol = pyscf.M(atom=self.atom, basis=self.basis)
        mol.UKS(xc=self.xc).sOPT().kernel()


class TestTSOPT:
    def setup_method(self, method):
        self.atom = """
N           0.00083273        0.00057666       -0.00069428
H           1.00555992       -0.00091855       -0.00131155
H          -0.50382419        0.86943925        0.00002190
H          -0.50256846       -0.86909736        0.00198393
"""
        self.basis = "sto-3g"
        self.xc = "LDA"

    def test_rhf(self):
        mol = pyscf.M(atom=self.atom, basis=self.basis)
        mol.RHF().sOPT(ts=True).kernel()

    def test_rks(self):
        mol = pyscf.M(atom=self.atom, basis=self.basis)
        mol.RKS(xc=self.xc).sOPT(ts=True).kernel()

    def test_uhf(self):
        mol = pyscf.M(atom=self.atom, basis=self.basis)
        mol.UHF().sOPT(ts=True).kernel()

    def test_uks(self):
        mol = pyscf.M(atom=self.atom, basis=self.basis)
        mol.UKS(xc=self.xc).sOPT(ts=True).kernel()


class TestTSSearch:
    def setup_method(self, method):
        self.product_xyzs = """4

  N    -0.00000000     0.00000000     0.07558051
  H     0.94112550     0.00000000    -0.35009807
  H    -0.47056275     0.81503859    -0.35009807
  H    -0.47056275    -0.81503859    -0.35009807
"""

        self.atom = """
  N    -0.00000000     0.00000000    -0.07558051
  H     0.94112550     0.00000000     0.35009807
  H    -0.47056275     0.81503859     0.35009807
  H    -0.47056275    -0.81503859     0.35009807
"""
        self.basis = "sto-3g"
        self.xc = "LDA"

    def test_rhf(self):
        mol = pyscf.M(atom=self.atom, basis=self.basis)
        mol.RHF().sOPT(ts=True, product_xyzs=self.product_xyzs).kernel()

    def test_rks(self):
        mol = pyscf.M(atom=self.atom, basis=self.basis)
        mol.RKS(xc=self.xc).sOPT(
            ts=True, product_xyzs=self.product_xyzs).kernel()

    def test_uhf(self):
        mol = pyscf.M(atom=self.atom, basis=self.basis)
        mol.UHF().sOPT(ts=True, product_xyzs=self.product_xyzs).kernel()

    def test_uks(self):
        mol = pyscf.M(atom=self.atom, basis=self.basis)
        mol.UKS(xc=self.xc).sOPT(
            ts=True, product_xyzs=self.product_xyzs).kernel()



# Copyright 2014-2022 The PySCF Developers, 2025-2026 Sidereus-AI ltd. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");


from pyscf import hessian, lib, scf
from .rhf import Infrared, get_h2ao_dipderiv, lib

import numpy as np


def proc_hessian_(mf_hess, mo_energy=None, mo_coeff=None, mo_occ=None, h1ao_grad=None):
    mf = mf_hess.base

    mo_energy = mo_energy if mo_energy else mf.mo_energy
    mo_coeff = mo_coeff if mo_coeff else mf.mo_coeff
    mo_occ = mo_occ if mo_occ else mf.mo_occ
    h1ao_grad = h1ao_grad if h1ao_grad else mf_hess.make_h1(mo_coeff, mo_occ)

    moao1_grad, mo_e1_grad = mf_hess.solve_mo1(
        mo_energy, mo_coeff, mo_occ, h1ao_grad)

    hess_elec = mf_hess.hess_elec(
        mo1=moao1_grad, mo_e1=mo_e1_grad, h1ao=h1ao_grad)
    hess_nuc = mf_hess.hess_nuc()
    mf_hess.de = hess_elec + hess_nuc

    mo1_grad = [None] * 2
    for s in (0, 1):
        mo1_grad[s] = lib.einsum(
            "up, uv, Axvi -> Axpi", mo_coeff[s], mf.get_ovlp(), moao1_grad[s])

    return mf_hess, h1ao_grad, mo1_grad, mo_e1_grad


def kernel_dipderiv(mf_ir):

    if mf_ir.mf_hess is None:
        mf_ir.mf_hess = mf_ir.hess_cls(mf_ir.base)
        # mf_ir.mf_hess.grid_response = True  # in case mf_hess is dft method, but seems no effect?
    if mf_ir.mf_hess.de is None \
            or np.abs(mf_ir.mf_hess.de).sum() < 1e-14 \
            or mf_ir._mo1_grad is NotImplemented:
        mf_hess, h1ao_grad, mo1_grad, mo_e1_grad = proc_hessian_(mf_ir.mf_hess)
        mf_ir._h1ao_grad = h1ao_grad
        mf_ir._mo1_grad = mo1_grad
        mf_ir._mo_e1_grad = mo_e1_grad
    mo1_grad = mf_ir._mo1_grad

    if mf_ir._h1_dip is NotImplemented:
        mf = mf_ir.base
        mo_coeff = mf.mo_coeff
        mo_occ = mf.mo_occ
        int_r = mf.mol.intor_symmetric("int1e_r")
        h1_dip = [None] * 2
        for s in (0, 1):
            h1_dip[s] = lib.einsum(
                "tuv, up, vi-> tpi", int_r, mo_coeff[s], mo_coeff[s][:, mo_occ[s] > 0])
        mf_ir._h1_dip = h1_dip
    h1_dip = mf_ir._h1_dip

    if mf_ir._h2ao_dipderiv is NotImplemented:
        mf_ir._h2ao_dipderiv = get_h2ao_dipderiv(mf_ir.base)
    h2ao_dipderiv = mf_ir._h2ao_dipderiv

    mol = mf_ir.mol
    natm = mol.natm
    de = np.zeros((natm, 3, 3))
    for a in range(natm):
        de[a] = np.eye(3) * mol.atom_charge(a)
    de += np.einsum("Axtuv, suv -> Axt", h2ao_dipderiv, mf_ir.base.make_rdm1())
    for s in (0, 1):
        de -= 2 * np.einsum("tpi, Axpi -> Axt", h1_dip[s], mo1_grad[s])

    mf_ir.de = de
    return de


class Infrared(Infrared):
    hess_cls = hessian.uhf.Hessian
    kernel_dipderiv = kernel_dipderiv


scf.uhf.UHF.sIR = lib.class_as_method(Infrared)



# Copyright 2014-2022 The PySCF Developers, 2025-2026 Sidereus-AI ltd. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");

from .rhf import Raman
from pyscf import lib, scf

class Raman(Raman):
    pass

scf.uhf.UHF.sRAMAN = lib.class_as_method(Raman)
